unit UnitPaste;
{
    Purpose:
        Move the Pasting logic into it's own class. The FrmMainPopup unit
        was getting a little bloated.
    NOTES;
        Use EmptyClipboard (or Clipboard.Clear) before pasting - this marks
        clips as "ours" so the ClipboardManager can ignore them

    Updates:
        Don't use Simple Text version when pasting CF_DIBs
        --------------------
        Pasting complex items will now also include a plain text (for programs
        that don't understand what, for example, CF_HTML is)
        --------------------
        New option to alter the clipboard (and trigger the Clipboard Manager)
        Win9x Mimic characters shift keys didn't work
        -----------
        Clear the clipboard before pasting
        This is the only place an item may be placed on the clipboard

        ------
        Another fix for SHIFT+Insert on Win2k
        --------------------
        - New support for pasting with keystrokes embedded in the text
        - Fix for SHIFT+Insert on Win2k
}
interface
uses  INIFiles, UnitFrmHotkey;

type TPasteMethod = (
    PASTE_CTRL_V=0,
    PASTE_SHIFT_INS=1,
    PASTE_MIMIC=2,
    PASTE_CLIPBOARD=3,
    PASTE_DEFAULT=4,
    PASTE_DISABLED=5
);

type TPaste = class(TObject)
    private
        {Configurations for pasting method}
        UseKeyboardMimic,
        UsePastingSI,
        UsePastingCV,
        UseClipboardOnly,

        {single time use overrides for pasting}
        PasteCVOnce,
        PasteSIOnce,
        ClipboardOnlyOnce,
        KeyboardMimicOnce : boolean;

        EXEPasteList : THashedStringList;

        procedure ClearMethods;
    public
        constructor Create;
        destructor Destroy; override;

        procedure SendKestroke(hk : TKeystrokeData);
        procedure SendTextWithKeystrokes(s : string);
        //procedure PlaceOnClipboardDontBypassClipboardManager(s : string);
        procedure SendSHIFT_INSERT;
        procedure SendCTRL_V;
        procedure SendCTRL_C;
        procedure SendTAB;
        procedure SendENTER;
        procedure SendINSERT;
        procedure SendDELETE;
        procedure SendBACKSPACE;
        procedure SendHOME;
        procedure SendEND;
        procedure SendUP;
        procedure SendDOWN;
        procedure SendLEFT;
        procedure SendRIGHT;
        procedure SendESC;
        procedure SendSPACE;
        procedure SendCTRL_SHIFT_LEFT;
        procedure SendCTRL_RIGHT;
        {Used by FrmCONFIG to set pasting method}
        procedure SetClipboardOnly;
        function GetClipboardOnly : boolean;
        procedure SetMimicTyping;
        procedure SetUsePaste;
        procedure SetUsePasteSI;

        {Used by FrmMainPopup}
        procedure SetClipboardOnlyOnce;
        procedure SetKeyboardMimicOnce;
        function GetKeyboardMimicOnce : boolean;
        procedure SetUsePasteCVOnce;
        procedure SetUsePasteSIOnce;
        procedure ClearOnceFlags;

        procedure AssignPaste(EXEName : string; method : TPasteMethod);
        function GetPasteMethod(EXEName : string) : TPasteMethod;
        function GetDefaultPasteMethod : TPasteMethod;
end;

var Paste : TPaste;

implementation

uses Windows, SysUtils, StrUtils, clipbrd,
    UnitMisc, UnitToken, Forms, Dialogs, UnitFrmGestures;

const  EXEPASTE_FILE = 'exepaste.ini';
{ TPaste }



//
// Either send text S or use ClipItem if specificed
//

//
// used by SendText and OtherItemClickEvent to simulate the user
// pressing CTRL+V to paste
//
procedure TPaste.SendCTRL_V;
var w : word;
begin
    UnitMisc.AppendLog('sending CTRL+V');

    // clear any phantom keystrokes
    keybd_event(VK_SHIFT, VkKeyScan(char(VK_SHIFT)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    sleep(10);

    // press CTRL, press V, release V, release CTRL
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), 0, 0);
    sleep(10);
            w := VkKeyScan('V');
            keybd_event(lo(w), w, 0, 0);
            sleep(50);

            keybd_event(lo(w), w, KEYEVENTF_KEYUP, 0);
            sleep(50);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent CTRL+V');
end;


procedure TPaste.SendCTRL_C;
var w : word;
begin
    UnitMisc.AppendLog('sending CTRL+C');

    // clear any phantom keystrokes
    keybd_event(VK_SHIFT, VkKeyScan(char(VK_SHIFT)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    sleep(10);

    // press CTRL, press V, release V, release CTRL
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), 0, 0);
    sleep(10);
            w := VkKeyScan('C');
            keybd_event(lo(w), w, 0, 0);
            sleep(10);

            keybd_event(lo(w), w, KEYEVENTF_KEYUP, 0);
            sleep(10);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    //sleep(10);
    UnitMisc.AppendLog('sent CTRL+C');
end;


procedure TPaste.SendCTRL_SHIFT_LEFT;
var w : word;
begin
     // clear any phantom keystrokes
    keybd_event(VK_SHIFT, VkKeyScan(char(VK_SHIFT)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    sleep(10);

    // press CTRL, press V, release V, release CTRL
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), 0, 0);
    sleep(1);
    keybd_event(VK_SHIFT, Lo(MapVirtualKey(VK_SHIFT,0)),0,0);  // win9x
    sleep(10);
            //w := VkKeyScan('C');
            w := VK_LEFT;
            keybd_event(lo(w), w, KEYEVENTF_EXTENDEDKEY, 0);
            sleep(10);

            keybd_event(lo(w), w, KEYEVENTF_EXTENDEDKEY or KEYEVENTF_KEYUP, 0);
            sleep(10);
    keybd_event(VK_SHIFT, Lo(MapVirtualKey(VK_SHIFT,0)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);

    //sleep(10);
    UnitMisc.AppendLog('sent CTRL+SHIFT+LEFT');
end;

procedure TPaste.SendKestroke(hk: TKeystrokeData);
var w : word;
var flags : cardinal;
begin

    // unpress mod keys
   { keybd_event(VK_SHIFT, VkKeyScan(char(VK_SHIFT)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_MENU , VkKeyScan(char(VK_MENU)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_LWIN, VkKeyScan(char(VK_MENU)), KEYEVENTF_KEYUP, 0);}
    sleep(10);



    // simulate modifiers down
    if hk.win then
        keybd_event(VK_LWIN, VkKeyScan(char(VK_LWIN)), 0, 0);
    sleep(10);
    if hk.alt then
        keybd_event(VK_MENU, VkKeyScan(char(VK_MENU)), 0, 0);
    sleep(10);
    if hk.ctrl then
        keybd_event(VK_CONTROL , VkKeyScan(char(VK_CONTROL)), 0, 0);
    sleep(10);
    if hk.shft  then
        keybd_event(VK_SHIFT, Lo(MapVirtualKey(VK_SHIFT,0)),0,0);  // win9x
    sleep(10);

            // main key press
            //w := VkKeyScan('C');
            w := hk.key;

            flags := 0;
            if (hk.key >= ord('A')) and (hk.key <= ord('Z')) then begin
                w := VKKeyScan(char(lo(w)));
            end;
           if hk.key in [VK_LEFT, VK_RIGHT, VK_HOME, VK_END, VK_NEXT, VK_PRIOR ] then
                flags := KEYEVENTF_EXTENDEDKEY;

            keybd_event(lo(w), w, flags, 0);
            sleep(10);

            keybd_event(lo(w), w, flags or KEYEVENTF_KEYUP, 0);
            sleep(10);
    //keybd_event(VK_SHIFT, Lo(MapVirtualKey(VK_SHIFT,0)), KEYEVENTF_KEYUP, 0);
    //keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);

    // release modifiers
    if hk.shft then
        keybd_event(VK_SHIFT, VkKeyScan(char(VK_SHIFT)), KEYEVENTF_KEYUP, 0);
    if hk.ctrl then
        keybd_event(VK_CONTROL, VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    if hk.alt then
        keybd_event(VK_MENU, VkKeyScan(char(VK_MENU)), KEYEVENTF_KEYUP, 0);
    if hk.win then
        keybd_event(VK_LWIN, VkKeyScan(char(VK_LWIN)), KEYEVENTF_KEYUP, 0);

    //sleep(10);
    UnitMisc.AppendLog('sent Kestroke');

end;


procedure TPaste.SendCTRL_RIGHT;
var w : word;
begin
     // clear any phantom keystrokes
    keybd_event(VK_SHIFT, VkKeyScan(char(VK_SHIFT)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    sleep(10);

    // press CTRL, press V, release V, release CTRL
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), 0, 0);
    sleep(1);
    //keybd_event(VK_SHIFT, Lo(MapVirtualKey(VK_SHIFT,0)),0,0);  // win9x
    sleep(10);
            //w := VkKeyScan('C');
            w := VK_RIGHT;
            keybd_event(lo(w), w, KEYEVENTF_EXTENDEDKEY, 0);
            sleep(10);

            keybd_event(lo(w), w, KEYEVENTF_EXTENDEDKEY or KEYEVENTF_KEYUP, 0);
            sleep(10);
    //keybd_event(VK_SHIFT, Lo(MapVirtualKey(VK_SHIFT,0)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);

    //sleep(10);
    UnitMisc.AppendLog('sent CTRL+SHIFT+LEFT');

end;

procedure TPaste.SendSHIFT_INSERT;
begin
    UnitMisc.AppendLog('sending SHIFT+INSERT');

    // clear any phantom keystrokes
    keybd_event(VK_SHIFT, VkKeyScan(char(VK_SHIFT)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    sleep(10);

    // press SHIFT, press INSERT, release INSERT, release SHIFT

    // Label VK_INSERT as extended, othewise it freaks out when combined with SHIFT

    keybd_event(VK_SHIFT, Lo(MapVirtualKey(VK_SHIFT,0)),0,0);  // win9x
    sleep(1);
            keybd_event(VK_INSERT, Lo(MapVirtualKey(VK_INSERT,0)),KEYEVENTF_EXTENDEDKEY, 0);
            sleep(1);

            keybd_event(VK_INSERT, Lo(MapVirtualKey(VK_INSERT,0)),KEYEVENTF_EXTENDEDKEY or KEYEVENTF_KEYUP,0);
            sleep(1);

    keybd_event(VK_SHIFT, Lo(MapVirtualKey(VK_SHIFT,0)), KEYEVENTF_KEYUP, 0);
    sleep(5);
    UnitMisc.AppendLog('sent SHIFT+INSERT');
end;

procedure TPaste.SendTAB;
begin
    UnitMisc.AppendLog('sending tab');
    keybd_event(VK_TAB , 0, 0, 0);
    sleep(30);

    keybd_event(VK_TAB, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent tab');
end;
procedure TPaste.SendDELETE;
begin
    UnitMisc.AppendLog('sending del');
    keybd_event(VK_DELETE, 0, 0, 0);
    sleep(30);

    keybd_event(VK_DELETE, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent del');

end;
procedure TPaste.SendINSERT;
begin
    UnitMisc.AppendLog('sending INSERT');
    keybd_event(VK_INSERT, 0, 0, 0);
    sleep(30);

    keybd_event(VK_INSERT, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent INSERT');
end;



procedure TPaste.SendENTER;
begin
    UnitMisc.AppendLog('sending insert');
    keybd_event(VK_RETURN, MapVirtualKey(VK_RETURN,0), 0, 0);
    sleep(30);

    keybd_event(VK_RETURN, MapVirtualKey(VK_RETURN,0), KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent insert');
end;
procedure TPaste.SendSPACE;
begin
    UnitMisc.AppendLog('sending tab');
    keybd_event(VK_SPACE , 0, 0, 0);
    sleep(30);

    keybd_event(VK_SPACE, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sending space');
end;





procedure TPaste.ClearMethods;
begin
    self.UseKeyboardMimic := false;
    self.UsePastingSI := false;
    self.UsePastingCV := false;
    self.UseClipboardOnly := false;
end;


procedure TPaste.SetClipboardOnly;
begin
    self.ClearMethods;
    self.UseClipboardOnly := true;
end;
procedure TPaste.SetMimicTyping;
begin
    self.ClearMethods;
    self.UseKeyboardMimic := true;
end;
procedure TPaste.SetUsePaste;
begin
    self.ClearMethods;
    self.UsePastingCV := true;
end;
procedure TPaste.SetUsePasteSI;
begin
    self.ClearMethods;
    self.UsePastingSI := true;
end;



procedure TPaste.SetKeyboardMimicOnce;
begin
    self.KeyboardMimicOnce := true;
end;
procedure TPaste.SetUsePasteCVOnce;
begin
    self.PasteCVOnce := true;
end;
procedure TPaste.SetUsePasteSIOnce;
begin
    self.PasteSIOnce := true;
end;
procedure TPaste.SetClipboardOnlyOnce;
begin
    self.ClipboardOnlyOnce := true;
end;

function TPaste.GetClipboardOnly: boolean;
begin
    result := self.UseClipboardOnly;
end;



{
//
// Not used, but may need to be refered to later
//
procedure TFrmMainPopup.PlaceTextOnClipboard(s : string);
    procedure SetBuffer(format: Word; var buffer; Size: Integer);
    var Data: THandle;
        DataPtr: Pointer;
    begin
        clipboard.Open;
        try
            Data := Windows.GlobalAlloc(GMEM_MOVEABLE+GMEM_DDESHARE, Size);
            try
                DataPtr := GlobalLock(Data);
                try
                    Move(Buffer, DataPtr^, Size);
                    clipboard.Clear;
                    Windows.SetClipboardData(Format, Data);
                finally
                    Windows.GlobalUnlock(Data);
                end;
            except
                Windows.GlobalFree(Data);
                raise;
            end;
        finally
            clipboard.Close;
        end;
    end;

begin

    s := s + #0 + 'ArsClip' + #0;
    SetBuffer(CF_TEXT, PChar(s)^, length(s) + 1);
end;
}



function TPaste.GetKeyboardMimicOnce: boolean;
begin
    result := self.UseKeyboardMimic;
end;

type TKeyProc = procedure;

procedure TPaste.SendTextWithKeystrokes(s: string);
var str : string;
    key, oldkey : string;
    datestr : string;
    clip : string;
const SLEEPMS = 60;

    procedure PreKey;
    begin
        //self.SendText(str);
        sleep(SLEEPMS);
    end;
    procedure PostKey;
    begin
        //Application.ProcessMessages;
        str := '';
    end;
begin
    // key rid of '[KEYS]'
    s := rightstr(s,length(s) - 6);

    if pos('[CLIPBRD]',s) <> 0 then begin
        clip := Clipboard.AsText;
    end;
    str := '';
    while (s <> '') do begin
        if (pos('[',s) <> 0) and (pos(']',s) <> 0) then begin
            str := str + TokenString(s, '[');
            // add everything upto the Bracket
            if (s <> '') then begin


                // allow a nested lookup for:
                // find the first "[ ]" without and "[" in between
                while (s <> '') and
                    (pos('[',s) <> 0) and
                    (pos('[',s) < pos(']',s)) do begin
                    str := str + '[' + TokenString(s, '[');
                end;

                // extract KEYSTROKEHERE]
                // from the text
                key := TokenString(s, ']');
                oldkey := key;
                key := UpperCase(key);




                if (key = 'TAB') then begin
                    prekey; self.SendTAB; PostKey;
                end else if (key = 'ENTER') then begin
                    prekey; self.SendENTER; PostKey;
                end else if (key = 'DEL') then begin
                    prekey; self.SendDELETE; PostKey;
                end else if (key = 'HOME') then begin
                    prekey; self.SendHOME; PostKey;
                end else if (key = 'BACK') then begin
                    prekey; self.SendBACKSPACE; PostKey;
                end else if (key = 'END') then begin
                    prekey; self.SendEND; PostKey;
                end else if (key = 'INS') then begin
                    prekey; self.SendINSERT; PostKey;
                end else if (key = 'UP') then begin
                    prekey; self.SendUP;PostKey;
                end else if (key = 'DOWN') then begin
                    prekey; self.SendDOWN; PostKey;
                end else if (key = 'LEFT') then begin
                    prekey;self.SendLEFT; PostKey;
                end else if (key = 'RIGHT') then begin
                   prekey; self.SendRIGHT; PostKey;
                end else if (key = 'SPACE') then begin
                   prekey; self.SendSPACE; PostKey;
                end else if (key = 'WAIT') then begin
                    sleep(100);
                end else if AnsiStartsText('DATE',key) then begin
                    // extract mm/dd/yy
                    // from DATE="mm/dd/yy"
                    TokenString(key,'"');
                    datestr := TokenString(key, '"');
                    str := str + FormatDateTime(datestr,now)
                end else if AnsiStartsText('CLEAR', key) then begin
                    prekey; clipboard.Clear; postkey;
                end else if (key = 'CLIPBRD') then begin
                        str := str + clip;
                end else if (key = 'RUN') then begin
                    UnitMisc.ShellExecute(FrmGestures.handle, s);
                    // cut everything or cut up to the next keystroke
                    TokenString(s, '[');
                    if s <> '' then begin
                        s := '[' + s;
                    end;
                end else begin
                    // not a key, just add it to they string to
                    // send buffer
                    str := str + '['+ oldkey + ']';
                end;
            end;
        end else begin
            str := str +s;
            s := '';
        end;
    end;
    
    if (s = '') then begin
        if (str <> '') then begin
            // TODO - fix this
            //self.SendText(str);
        end;
    end;

end;





procedure TPaste.SendBACKSPACE;
begin
    UnitMisc.AppendLog('sending back');
    keybd_event(VK_BACK , 0, 0, 0);
    sleep(30);

    keybd_event(VK_BACK, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent back');
end;

procedure TPaste.SendEND;
begin
    UnitMisc.AppendLog('sending end');
    keybd_event(VK_END  , 0, 0, 0);
    sleep(30);

    keybd_event(VK_END, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent end');
end;

procedure TPaste.SendHOME;
begin
    UnitMisc.AppendLog('sending VK_HOME');
    keybd_event(VK_HOME  , 0, 0, 0);
    sleep(30);

    keybd_event(VK_HOME, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent VK_HOME');
end;

procedure TPaste.SendDOWN;
begin
     UnitMisc.AppendLog('sending VK_DOWN');
    keybd_event(VK_DOWN   , 0, 0, 0);
    sleep(30);

    keybd_event(VK_DOWN, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent VK_DOWN');
end;

procedure TPaste.SendLEFT;
begin
  UnitMisc.AppendLog('sending VK_LEFT');
    keybd_event(VK_LEFT   , 0, 0, 0);
    sleep(30);

    keybd_event(VK_LEFT, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent VK_LEFT');
end;

procedure TPaste.SendRIGHT;
begin
   UnitMisc.AppendLog('sending VK_RIGHT');
    keybd_event(VK_RIGHT   , 0, 0, 0);
    sleep(30);

    keybd_event(VK_RIGHT, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent VK_RIGHT');
end;

procedure TPaste.SendUP;
begin
   UnitMisc.AppendLog('sending VK_UP');
    keybd_event(VK_UP   , 0, 0, 0);
    sleep(30);

    keybd_event(VK_UP, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent VK_UP');
end;





procedure TPaste.ClearOnceFlags;
begin
    PasteCVOnce := false;
    PasteSIOnce := false;
    ClipboardOnlyOnce := false;
    KeyboardMimicOnce := false;
end;


//
// Assigned Paste Methods
//
constructor TPaste.Create;
var s : string;
begin
    self.EXEPasteList := THashedStringList.Create;
    s := UnitMisc.GetAppPath;
    s := s + EXEPASTE_FILE;
    if FileExists(s) then begin
        self.EXEPasteList.LoadFromFile(s);
    end;
end;

destructor TPaste.Destroy;
var s : string;
begin
    s := UnitMisc.GetAppPath;
    s := s + EXEPASTE_FILE;
    self.EXEPasteList.SaveToFile(s);
    MyFree(EXEPasteList);
end;


procedure TPaste.AssignPaste(EXEName: string; method: TPasteMethod);
begin
    EXEPasteList.Values[EXEName] := IntToStr(Integer(method));
end;

function TPaste.GetPasteMethod(EXEName : string) : TPasteMethod;
var s : string;
begin
    result := PASTE_DEFAULT;
    s := EXEPasteList.Values[EXEName];

    if (s <> '') then begin
        result := TPasteMethod(StrToInt(s));
    end;
end;


function TPaste.GetDefaultPasteMethod: TPasteMethod;
begin
    result := PASTE_DEFAULT; // this case should never fire

    if self.UseKeyboardMimic then begin
        result := PASTE_MIMIC;
    end else if self.UsePastingSI then begin
        result := PASTE_SHIFT_INS;
    end else if self.UsePastingCV then begin
        result := PASTE_CTRL_V;
    end else if self.UseClipboardOnly then begin
        result := PASTE_CLIPBOARD
    end;
end;

procedure TPaste.SendESC;
begin
    keybd_event(VK_ESCAPE  , 0, 0, 0);
    sleep(150);

    keybd_event(VK_ESCAPE, 0, KEYEVENTF_KEYUP, 0);
    sleep(30);
end;

initialization
begin
    UnitMisc.AppendLog('TPaste - creating');
    Paste := TPaste.Create;
end;
finalization
begin
    MyFree(Paste);
end;
end.
